#include "PMap.h"

// #####################################################################
/**
 * This function checks the cone condtion of a given set. 
 * @param[in] N - h set
*/
void checkConeCondition(PMap& pm, TripleSet &H, string message,int iterations)
{
	writeReport("Cone condition on " + message + ": ");
  GridSet grid(3);
  H.gridSet(grid,1,1);
  IMatrix DP = pm.derivative(*grid.begin(),grid.C,grid.r,iterations);

  IMatrix Q(2,2); Q[0][0] = 1; Q[1][1]=-1;
  IMatrix C(2,2);
  C.column(0) = H.Iunstable;
  C.column(1) = H.Istable;

  std::ofstream out;
  std::string filename = message+".dat"; 
  out.open(filename);
  out.precision(17);
  out << "quadratic form Q = " << Q << endl;
  out << "coordinates of the set C = " << C << endl;  
  DP = capd::matrixAlgorithms::krawczykInverse(C)*DP;
  out << "DP in coordinates of the set, i.e. V = C^{-1}*DP*C = " << DP << endl;  
  out << "The matrix V^T*Q*V-Q should be positive definite:\n";
  IMatrix V = Transpose(DP)*Q*DP-Q;
  out << "V = " << V << endl;
  out << "From Sylvester criterion we should have V(0,0)>0 and det(V)>0. We have\n";
  interval z;
  intersection(V[0][1],V[1][0],z);
  out << "V(0,0)=" << V[0][0] << endl;
  out << "det(V) = " << V[0][0]*V[1][1] - sqr(z) << endl;
  out.close();
  bool result = V[0][0] > 0 and (V[0][0]*V[1][1] - sqr(z) > 0);
  cout << (result ? "done" : "failed") << " (See file " << filename  << " for details)" << endl;
}


// #####################################################################
/**
 * This function checks if the condition holds true for a Poincare map of the set covered by a grid
 * @param[in] grid - grid of a domain
 * @param[in] Q,y - coordinate system in the image
 * @param[in] iterations - number of iterations of the Poincare map
 * @param[in] condition to be satisfied
 * @return true if condition satisfied
 */ 
template<class Condition>
bool checkCondition(PMap& pm, GridSet& grid, capd3::IMatrix Q, capd3::IVector y, int iterations, Condition condition, std::string message=""){
  GridSet::iterator b = grid.begin();
  std::list< std::pair<capd3::IVector,capd3::IVector> > unchecked;
  while(b!=grid.end()){
    unchecked.push_back(std::make_pair(*b,grid.r));
    ++b;
  }
  
  const int loop_limit = 400;
  
  while(!unchecked.empty() and unchecked.size()<loop_limit){
    auto elem = unchecked.front();
    unchecked.pop_front();
    IVector im = pm(elem.first,grid.C,elem.second,Q,y,iterations);
    if(!condition(im)){
      if(unchecked.size()+2>=loop_limit){
        writeReport("\n"+message+" (Inclusion not satisfied) ");
        std::cout << im << std::flush;
      }
      // subdivision of element
      capd3::IVector m = midVector(elem.second);
      capd3::IVector x1 = intervalHull(leftVector(elem.second),m);
      capd3::IVector x2 = intervalHull(rightVector(elem.second),m);
      capd3::IVector r1(3), r2(3);
      split(x1,r1);
      split(x2,r2);
      unchecked.push_back(std::make_pair(elem.first + grid.C*x1,r1));
      unchecked.push_back(std::make_pair(elem.first + grid.C*x2,r2));
    }
  }
  return unchecked.empty();
}

// #####################################################################
/**
 * This function checks covering relation between two sets
 * @param[in] N,M - h-sets
 * @param[in] iterations - number of iterations of the Poincare map
 * @param[in] orientation - positive if we do not change the orientation
 * @return true if the covering relation is validated
 */ 
bool checkCovering(PMap& pm, TripleSet& N, TripleSet& M, string message, int orientation, int iterations){
  writeReport(message);
  bool result = true;
  GridSet grid(3);
  capd3::IMatrix Q = pm.embed(M.Iunstable,M.Istable);
  Q = capd::matrixAlgorithms::krawczykInverse(Q);
  capd3::IVector y = pm.embed(M.Icenter);

  auto left = [](const IVector& x){return x[0]<-1.;};
  auto right = [](const IVector& x){return x[0]>1.;};
  auto across = [](const IVector& x){return capd::abs(x[1])<1. or x[0]>1. or x[0]<-1.;};

  const int initHorizontalGridSize = 5;
  const int initVerticalGridSize = 5;
  N.gridTopEdge(grid,initHorizontalGridSize);
  result = result and checkCondition(pm,grid,Q,y,iterations,across,"top ");
  N.gridBottomEdge(grid,initHorizontalGridSize);
  result = result and checkCondition(pm,grid,Q,y,iterations,across,"bottom ");
  N.gridRightEdge(grid,initVerticalGridSize);
  result = result and checkCondition(pm,grid,Q,y,iterations,orientation > 0 ? right : left,"right ");
  N.gridLeftEdge(grid,initVerticalGridSize);
  result = result and checkCondition(pm,grid,Q,y,iterations,orientation > 0 ? left : right,"left ");
  
  // existence of Poincare map
  // subdivision (5x5) for all sets
  // no restrictions on the image
  auto alwaysTrue = [](const IVector&) { return true; };
  N.gridSet(grid,5,5);
  checkCondition(pm,grid,Q,y,iterations,alwaysTrue,"exists ");
  writeReport(result ? ": done\n" : ": failed\n");
  
  return result;
}

// #####################################################################
/**
 * This function checks if the condition holds true for a Poincare map of the set covered by a grid
 * @param[in] grid - grid of a domain
 * @param[in] Q,y - coordinate system in the image
 * @param[in] T - time of integration
 * @param[in] condition to be satisfied
 * @return true if condition satisfied
 */ 
template<class Condition>
bool checkCondition(PMap& pm, GridSet& grid, capd3::IMatrix Q, capd3::IVector y, Condition condition, std::string message, interval T){
  GridSet::iterator b = grid.begin();
  while(b!=grid.end()){
    capd3::IVector im = pm(*b,grid.C,grid.r,Q,y,T);
    if(!condition(im)){
        writeReport("\n"+message+" (Inclusion not satisfied) ");
        std::cout << im << std::flush;
        return false;
    }
    ++b;
  }
  return true;
}

// #####################################################################
/**
 * This function checks covering relation 
 * between two-dimensional h-set and three-dimensional h-set
 * @param[in] N,M - h-sets 2-dim and 3-dim, respectively
 * @param[in] T - time of integration
 * @return throws and exception when the condition is not satisfied
 */ 
bool checkCovering(PMap& pm, TripleSet& N, HSet3D& M, string message, interval T){
  writeReport(message);
  bool result = true;
  GridSet grid(3);
  capd3::IMatrix Q = capd::matrixAlgorithms::krawczykInverse(M.IM);
  capd3::IVector y = M.Icenter;

  auto left = [](const capd3::IVector& x){return x[0]<-1.;};
  auto right = [](const capd3::IVector& x){return x[0]>1.;};
  auto across = [](const capd3::IVector& x){return (capd::abs(x[1])<1. and capd::abs(x[2])<1.) or x[0]>1. or x[0]<-1.;};

  const int initHorizontalGridSize = 6;
  const int initVerticalGridSize = 3;
  N.gridSet(grid,initHorizontalGridSize,initVerticalGridSize);
  result = result and checkCondition(pm,grid,Q,y,across,"across ",T);
  N.gridRightEdge(grid,initVerticalGridSize);
  result = result and checkCondition(pm,grid,Q,y,right,"right ",T);
  N.gridLeftEdge(grid,initVerticalGridSize);
  result = result and checkCondition(pm,grid,Q,y,left,"left ",T);
  
  writeReport(result ? ": done\n" : ": failed\n");
  
  return result;
}

// #####################################################################
/**
 * This function checks covering relation 
 * between two three-dimensional h-sets
 * @param[in] N,M - h-sets 
 * @param[in] T - time of integration
 * @return throws and exception when the condition is not satisfied
 */
bool checkCovering(PMap& pm, HSet3D& N, HSet3D& M, string message, interval T){
  writeReport(message);
  bool result = true;
  GridSet grid(3);
  capd3::IMatrix Q = capd::matrixAlgorithms::krawczykInverse(M.IM);
  capd3::IVector y = M.Icenter;

  auto left = [](const capd3::IVector& x){return x[0]<-1.;};
  auto right = [](const capd3::IVector& x){return x[0]>1.;};
  auto across = [](const capd3::IVector& x){return (capd::abs(x[1])<1. and capd::abs(x[2])<1.) or x[0]>1. or x[0]<-1.;};

  int g = 5;
  N.gridWall(grid,g,g,1);
  result = result and checkCondition(pm,grid,Q,y,across,"across ",T);
  N.gridWall(grid,g,g,2);
  result = result and checkCondition(pm,grid,Q,y,across,"across ",T);

  N.gridWall(grid,1,1,0,1);
  result = result and checkCondition(pm,grid,Q,y,right,"right ",T);
  N.gridWall(grid,1,1,0,-1);
  result = result and checkCondition(pm,grid,Q,y,left,"left ",T);
  
  writeReport(result ? ": done\n" : ": failed\n");
  
  return result;
}

