#include "main.h"
#include "PMap.h"

void writeReport(std::string s){
	std::cout << s << std::flush;
}

TripleSet N[5];
TripleSet H1, H2, G1, G2, M;
HSet3D H;

//###########################################################//
/**
 * Here we call generic algorithm (see algorithms.cpp) 
 * that checks if an h-sets A covers B by Poincare map. 
 * The sign 1 or -1 means that we change or not the orientation.
 */ 
void checkCoveringRelations_N1_N5(PMap& pm)
{    
  writeReport("Covering relations for symbolic dynamics (JDE'2003).\n");
  checkCovering(pm, N[0], N[3], "N0=>N3", -1);
  checkCovering(pm, N[4], N[1], "N4=>N1", 1);
  checkCovering(pm, N[1], N[2], "N1=>N2", -1);
  checkCovering(pm, N[2], N[3], "N2=>N3", -1);
  checkCovering(pm, N[3], N[0], "N3=>N0", 1);
  std::cout << std::endl;
}

// #################################################################
/**
 * This function realizes shooting method for symmetric periodic orbit.
 * Since the Michelson system is reversible it is enough to check that
 * pi_x(P(0,y1))* pi_x(P(0,y_2)) < 0
 * 
 * Then if the Poincare map is defined on 0x[y1,y2] then there is 
 * y0\in [y1,y2] such that pi_x(P(0,y)) = 0 and the point (0,y) is a symetric periodic point. 
 */ 
void checkSymmetricPeriodicOrbit(PMap& pm, IVector u, std::string message)
{
  writeReport("Existence of symmetric periodic orbit " + message + ": ");
	IVector u1 = leftVector(H1.Icenter);
  IVector v1 = pm(u1,1); 
	IVector u2 = rightVector(H1.Icenter);
  IVector v2 = pm(u2,1); 

	bool result = (v1[0]*v2[0]<0);
  writeReport(result ? "done\n" : "failed\n");
  if(!result){
    std::cout << v1 << std::endl;
    std::cout << v2 << std::endl;
  }
}

//###########################################################//
/**
 * The cone conditions are checked on very small sets H1, H2 centered at S1, S2.
 * In order to obtain homoclinic and heteroclinic orbits we show that 
 * there are sets G1, G2 such that
 *  H1 => H1 => G1 => N0
 *  H2 => H2 => G2 => N1
 * These sets are self-symmetric. Therefore using backovering theory we obtain
 * homoclinic and heteroclinic orbits to both S1 and S2.
 */ 
void checkConeConditions_H1_H2(PMap& pm)
{
  writeReport("Covering relations for convergence to periodic orbits S1 and S2.\n");
  // 2 is period, -1 - orientation is changed
  checkCovering(pm, H1, H1, "H1=>H1", -1, 2);
  checkCovering(pm, H1, G1, "H1=>G1", -1, 2);
  checkCovering(pm, G1, N[0], "G1=>N0", -1, 2);

  // 4 is period, 1 - orientation is not changed
  checkCovering(pm, H2, H2, "H2=>H2", 1, 4);
  checkCovering(pm, H2, G2, "H2=>G2", 1, 4);
  checkCovering(pm, G2, N[1], "G2=>N1", 1, 4);

  checkConeCondition(pm, H1, "H1",2);
  checkConeCondition(pm, H2, "H2",4);
  std::cout << std::endl;
}

//###########################################################//

void checkDomainH(PMap& pm){
  GridSet grid(3);
  H.gridSet(grid,4,4,4);
  GridSet::iterator b = grid.begin();
  pm.tm.stopAfterStep(true);
  interval T = interval(14)/10;
  while(b!=grid.end()){
    capd3::C0HOTripletonSet s(*b,grid.C,grid.r);
    do{
      pm.tm(T,s);
      if(! (s.getLastEnclosure()[0] < 0.) ){
        std::cout << "Inclusion not satisfied\n";
        cout << s.getLastEnclosure() << std::endl;
        cout << "failed\n";
      }
    }while(!pm.tm.completed());
    ++b;
  }
  std::cout << "done\n";
}

/*----------------------- MAIN ------------------------------*/
int main(int argc, char* argv[])
{
  PMap pm;
	initSets();

	try{    
    // check covering relations for chaotic dynamics as describe in the paper
    // Chaos in the Kuramoto-Sivashinsky equations - a computer assisted proof
    // JDE'2003
    checkCoveringRelations_N1_N5(pm);

    // All the remainig computations are related to the paper 
    // Symmetric heteroclinic connections in the Michelson system - a computer assisted proof
    // SIADS'2005
    writeReport("\nComputations related to the paper SIADS'2005.\n\n");
    
    // validate existence of two symmetric periodic orbits
    checkSymmetricPeriodicOrbit(pm,H1.Icenter,"S1");
    checkSymmetricPeriodicOrbit(pm,H2.Icenter,"S2");
    
    // Check cone conditions on h-sets centered at these periodic orbits
    // Cone conditions are used in the proof of the existence of homoclinic and heteroclinic solutions.
    checkConeConditions_H1_H2(pm);

    // Check conditions that guarantee convergence to equilibrium point.
    writeReport("Covering relation from 2-dim set to 3-dim set by the flow translation with time T=6.4\n");
    checkCovering(pm, M, H, "M=>H",interval(64)/10);

    writeReport("Covering relation from 3-dim set to 3-dim set by the flow translation with time T=1.4\n");
    checkCovering(pm, H, H, "H=>H",interval(14)/10);

    writeReport("Check if phi([0,1.4],H) is in the half-space x<0: ");
    checkDomainH(pm);
	}
	catch(std::exception &e)
	{
		std::cout << "Exception caught: " << e.what() << std::endl;
	}

  return 0;
}

